from pathlib import Path

from pydantic import Field

from cogentic.orchestration.models.base import CogenticBaseModel


class CogenticArtifact(CogenticBaseModel):
    """Artifact for the cogentic system."""

    name: str = Field(description="Name of the artifact")
    description: str = Field(description="Description of the artifact")
    path: Path = Field(
        description="Path to the artifact. This should be a relative path to the current working directory."
    )
    test_name: str = Field(description="Name of the test which produced this artifact")
    team_member_name: str = Field(
        description="Name of the team member who produced this artifact"
    )



from autogen_agentchat.state import BaseGroupChatManagerState
from pydantic import Field

from cogentic.orchestration.models.ledger import CogenticProgressLedger
from cogentic.orchestration.models.plan import CogenticPlan


class CogenticState(BaseGroupChatManagerState):
    """State for the cogentic system."""

    question: str = Field(default="")
    plan: CogenticPlan | None = Field(default=None)
    ledger: CogenticProgressLedger | None = Field(default=None)
    total_turns: int = Field(default=0)
    stalls: int = Field(default=0)
    type: str = Field(default="CogenticState")



import logging
from typing import Callable, List

from autogen_agentchat import EVENT_LOGGER_NAME, TRACE_LOGGER_NAME
from autogen_agentchat.base import ChatAgent, TerminationCondition
from autogen_agentchat.teams._group_chat._base_group_chat import BaseGroupChat
from autogen_core import Component, ComponentModel
from autogen_core.models import ChatCompletionClient
from pydantic import BaseModel
from typing_extensions import Self

from cogentic.orchestration.orchestrator import CogenticOrchestrator
from cogentic.orchestration.prompts.prompts import FINAL_ANSWER_PROMPT

trace_logger = logging.getLogger(TRACE_LOGGER_NAME)
event_logger = logging.getLogger(EVENT_LOGGER_NAME)


class CogenticGroupChatConfig(BaseModel):
    """The declarative configuration for a CogenticGroupChat."""

    participants: List[ComponentModel]
    model_client: ComponentModel
    termination_condition: ComponentModel | None = None
    max_turns_total: int | None = None
    max_turns_per_hypothesis: int | None = None
    max_turns_per_test: int | None = None
    max_stalls: int
    final_answer_prompt: str


class CogenticGroupChat(BaseGroupChat, Component[CogenticGroupChatConfig]):
    """A team that runs a group chat with participants managed by the CogenticOrchestrator."""

    component_config_schema = CogenticGroupChatConfig
    component_provider_override = "autogen_agentchat.teams.CogenticGroupChat"

    def __init__(
        self,
        participants: List[ChatAgent],
        model_client: ChatCompletionClient,
        json_model_client: ChatCompletionClient | None = None,
        *,
        max_turns_total: int | None = 128,
        max_turns_per_hypothesis: int | None = 32,
        max_turns_per_test: int | None = 8,
        max_stalls: int = 3,
        final_answer_prompt: str = FINAL_ANSWER_PROMPT,
    ):
        """Initialize the CogenticGroupChat.

        Args:

            participants (List[ChatAgent]): The participants in the group chat.
            model_client (ChatCompletionClient): The model client to use for orchestration.
            json_model_client (ChatCompletionClient | None): The model client to use for JSON response type generation. Defaults to None, in which case model_client is used.
            max_turns_total (int | None): The maximum number of turns for the group chat. Defaults to 128.
            max_turns_per_hypothesis (int | None): The maximum number of turns per hypothesis. Defaults to 32.
            max_turns_per_test (int | None): The maximum number of turns per test. Defaults to 8.
            max_stalls (int): The maximum number of stalls before the group chat is terminated. Defaults to 3.
            final_answer_prompt (str): The prompt to use for the final answer. Defaults to FINAL_ANSWER_PROMPT.
        """
        super().__init__(
            participants,
            group_chat_manager_class=CogenticOrchestrator,
            termination_condition=None,
            max_turns=max_turns_total,
        )

        # Validate the participants.
        if len(participants) == 0:
            raise ValueError(
                "At least one participant is required for CogenticGroupChat."
            )
        self._model_client = model_client
        # JSON model client is used for getting a JSON-object out of a prompt
        if json_model_client is None:
            json_model_client = model_client
        self._json_model_client = json_model_client
        self._max_stalls = max_stalls
        self._max_turns_per_hypothesis = max_turns_per_hypothesis
        self._max_turns_per_test = max_turns_per_test
        self._final_answer_prompt = final_answer_prompt

    def _create_group_chat_manager_factory(
        self,
        group_topic_type: str,
        output_topic_type: str,
        participant_topic_types: List[str],
        participant_descriptions: List[str],
        termination_condition: TerminationCondition | None,
        max_turns: int | None,
    ) -> Callable[[], CogenticOrchestrator]:
        return lambda: CogenticOrchestrator(
            group_topic_type=group_topic_type,
            output_topic_type=output_topic_type,
            participant_topic_types=participant_topic_types,
            participant_descriptions=participant_descriptions,
            model_client=self._model_client,
            json_model_client=self._json_model_client,
            max_turns_total=max_turns,
            max_turns_per_hypothesis=self._max_turns_per_hypothesis,
            max_turns_per_test=self._max_turns_per_test,
            max_stalls=self._max_stalls,
            final_answer_prompt=self._final_answer_prompt,
        )

    def _to_config(self) -> CogenticGroupChatConfig:
        participants = [
            participant.dump_component() for participant in self._participants
        ]
        termination_condition = (
            self._termination_condition.dump_component()
            if self._termination_condition
            else None
        )
        return CogenticGroupChatConfig(
            participants=participants,
            model_client=self._model_client.dump_component(),
            termination_condition=termination_condition,
            max_turns_total=self._max_turns,
            max_turns_per_hypothesis=self._max_turns_per_hypothesis,
            max_stalls=self._max_stalls,
            final_answer_prompt=self._final_answer_prompt,
        )

    @classmethod
    def _from_config(cls, config: CogenticGroupChatConfig) -> Self:
        participants = [
            ChatAgent.load_component(participant) for participant in config.participants
        ]
        model_client = ChatCompletionClient.load_component(config.model_client)
        return cls(
            participants,
            model_client,
            max_turns_total=config.max_turns_total,
            max_turns_per_hypothesis=config.max_turns_per_hypothesis,
            max_turns_per_test=config.max_turns_per_test,
            max_stalls=config.max_stalls,
            final_answer_prompt=config.final_answer_prompt,
        )



from pydantic import Field

from cogentic.orchestration.models.base import CogenticBaseModel


class CogenticAction(CogenticBaseModel):
    goal: str = Field(
        description="Short summary of the goal of the action",
    )
    outcome: str = Field(
        description="Short summary of the outcome of the action",
    )
    test_name: str = Field(description="Name of the test this action was a part of")
    team_member_name: str = Field(
        description="Name of the team member who performed the action"
    )



from __future__ import annotations

from typing import Literal, Self, Type

from pydantic import Field, model_validator

from cogentic.orchestration.models.base import CogenticBaseModel
from cogentic.orchestration.models.hypothesis import (
    CogenticHypothesis,
    CogenticHypothesisState,
)
from cogentic.orchestration.models.plan import CogenticPlanState
from cogentic.orchestration.models.reasoning import (
    CogenticReasonedChoiceAnswer,
    CogenticReasonedStringAnswer,
)
from cogentic.orchestration.models.test import CogenticTest


class CogenticNextStep(CogenticBaseModel):
    """Next step for the cogentic system."""

    goal: CogenticReasonedStringAnswer = Field(
        description="What is the goal of our next action? Provide your reasoning along with your answer.",
    )
    next_speaker: CogenticReasonedChoiceAnswer = Field(
        description="Who are we communicating with? Provide your reasoning and select the team member to communicate with.",
    )
    instruction_or_question: CogenticReasonedStringAnswer = Field(
        description="What instruction or question are we giving this team member? Provide your reasoning, and phrase your answer as it you're speaking to them directly.",
    )

    @classmethod
    def with_speaker_choices(cls, choices: list[str]) -> Type["CogenticNextStep"]:
        """Create a new type from our class, where the next speaker is limited to a set of choices."""
        # Create the choice type with proper annotation
        next_speaker_type = CogenticReasonedChoiceAnswer[Literal[tuple(choices)]]

        return type(
            "CogenticNextStepWithSpeakerChoices",
            (cls,),
            {
                "__annotations__": {
                    "next_speaker": next_speaker_type,
                },
            },
        )


class CogenticFinalAnswer(CogenticBaseModel):
    """Final answer for the cogentic system."""

    result: str = Field(description="The result of our work")
    completed_by_team_members: bool = Field(
        description="Whether the answer was completed by team members or by yourself"
    )
    status: Literal["complete", "incomplete"] = Field(
        description="Whether we were able to fully answer the question"
    )
    failure_reason: str | None = Field(
        description="Reason for the status (if incomplete)"
    )


class CogenticHypothesisUpdate(CogenticBaseModel):
    hypothesis_state: CogenticReasonedChoiceAnswer[CogenticHypothesisState] = Field(
        description="The state of the hypothesis.",
    )
    new_tests: list[CogenticTest] = Field(
        description="New tests to be added to the hypothesis if you're marking the state as `unverified`",
    )

    @model_validator(mode="after")
    def validate_hypothesis_state(self) -> Self:
        """Validate the hypothesis state."""
        if self.hypothesis_state.answer == "unverified" and not self.new_tests:
            raise ValueError(
                "If you're marking the hypothesis as `unverified`, you must provide new tests."
            )
        return self


class CogenticPlanUpdate(CogenticBaseModel):
    """Plan update for the cogentic system."""

    plan_state: CogenticReasonedChoiceAnswer[CogenticPlanState] = Field(
        description="The state of the plan.",
    )
    new_hypotheses: list[CogenticHypothesis] = Field(
        description="New hypotheses to be added to the plan if you're marking the state as `in_progress`",
    )

    @model_validator(mode="after")
    def validate_plan_state(self) -> Self:
        """Validate the plan state."""
        if self.plan_state.answer == "in_progress" and not self.new_hypotheses:
            raise ValueError(
                "If you're marking the plan as `in_progress`, you must provide new hypotheses."
            )
        return self



from typing import Literal

from pydantic import Field

from cogentic.orchestration.models.base import CogenticBaseModel

CogenticTestState = Literal["complete", "incomplete", "abandoned"]


class CogenticTestTeamMemberPlan(CogenticBaseModel):
    """Role of the team member in a test."""

    name: str = Field(description="Name of the team member")
    action: str = Field(description="Action we plan this team member to take.")
    rationale: str = Field(
        description="Why we believe this team member will be able to perform this action, based on their description"
    )


class CogenticTest(CogenticBaseModel):
    """A test which is part of a hypothesis."""

    name: str = Field(description="Name of the test")
    description: str = Field(description="Description of the test")
    goal: str = Field(description="Goal of the test")
    state: CogenticTestState = Field(description="State of the test")
    plan: list[CogenticTestTeamMemberPlan] = Field(
        description="Plan for the test. This should include a list of team members and how we envision them solving the test"
    )
    result_summary: str | None = Field(description="A summary of the results")



import json

from pydantic import BaseModel


class CogenticBaseModel(BaseModel):
    def model_dump_markdown(
        self, title: str | None = None, title_level: int = 2, indent: int = 2
    ) -> str:
        """Dump the model as a markdown string."""
        if title:
            title = f"{'#' * title_level} {title}\n\n"
        else:
            title = ""
        return f"{title}```json\n{self.model_dump_json(indent=indent)}\n```"

    def model_dump_field_as_markdown(
        self,
        field_name: str,
        title: str | None = None,
        title_level: int = 2,
        indent: int = 2,
        collapse_field: bool = True,
    ) -> str:
        """Dump a field as a markdown string."""
        if title:
            title = f"{'#' * title_level} {title}\n\n"
        else:
            title = ""
        # If we want to collapse the field, we should pull it out from the object dump before dumping to a json string
        if collapse_field:
            dumped_field = self.model_dump(include={field_name})[field_name]
            # If the field is a pydantic model we should model_dump_json. Otherwise, we should just use json.dumps
            if isinstance(dumped_field, BaseModel):
                dumped_field = dumped_field.model_dump_json(indent=indent)
            else:
                dumped_field = json.dumps(dumped_field, indent=indent)
            return f"{title}```json\n{dumped_field}\n```"
        # Otherwise, just dump the whole object
        else:
            return f"{title}```json\n{self.model_dump_json(indent=indent, include={field_name})}\n```"



from cogentic.orchestration.prompts.prompts import (
    create_current_state_prompt,
    create_final_answer_prompt,
    create_initial_evidence_prompt,
    create_initial_hypotheses_prompt,
    create_next_step_prompt,
    create_persona_prompt,
    create_progress_ledger_prompt,
    create_summarize_result_prompt,
    create_update_hypothesis_on_stall_prompt,
    create_update_hypothesis_prompt,
    create_update_plan_on_stall_prompt,
    create_update_plan_prompt,
)

__all__ = [
    "create_final_answer_prompt",
    "create_current_state_prompt",
    "create_initial_evidence_prompt",
    "create_initial_hypotheses_prompt",
    "create_next_step_prompt",
    "create_summarize_result_prompt",
    "create_persona_prompt",
    "create_progress_ledger_prompt",
    "create_update_hypothesis_prompt",
    "create_update_hypothesis_on_stall_prompt",
    "create_update_plan_prompt",
    "create_update_plan_on_stall_prompt",
]



import logging
import re
from typing import Any, List, Mapping

from autogen_agentchat import TRACE_LOGGER_NAME
from autogen_agentchat.base import Response
from autogen_agentchat.messages import (
    AgentEvent,
    ChatMessage,
    HandoffMessage,
    MultiModalMessage,
    StopMessage,
    TextMessage,
    ToolCallExecutionEvent,
    ToolCallRequestEvent,
    ToolCallSummaryMessage,
)
from autogen_agentchat.teams._group_chat._base_group_chat_manager import (
    BaseGroupChatManager,
)
from autogen_agentchat.teams._group_chat._events import (
    GroupChatAgentResponse,
    GroupChatMessage,
    GroupChatRequestPublish,
    GroupChatReset,
    GroupChatStart,
    GroupChatTermination,
)
from autogen_agentchat.utils import content_to_str, remove_images
from autogen_core import (
    AgentId,
    CancellationToken,
    DefaultTopicId,
    MessageContext,
    event,
    rpc,
)
from autogen_core.models import (
    AssistantMessage,
    ChatCompletionClient,
    LLMMessage,
    SystemMessage,
    UserMessage,
)

from cogentic.orchestration.model_output import reason_and_output_model
from cogentic.orchestration.models.action import CogenticAction
from cogentic.orchestration.models.evidence import CogenticInitialEvidence
from cogentic.orchestration.models.hypothesis import CogenticInitialHypotheses
from cogentic.orchestration.models.ledger import CogenticProgressLedger
from cogentic.orchestration.models.orchestration import (
    CogenticFinalAnswer,
    CogenticHypothesisUpdate,
    CogenticNextStep,
    CogenticPlanUpdate,
)
from cogentic.orchestration.models.plan import CogenticPlan
from cogentic.orchestration.models.state import CogenticState
from cogentic.orchestration.prompts import (
    create_current_state_prompt,
    create_final_answer_prompt,
    create_initial_evidence_prompt,
    create_initial_hypotheses_prompt,
    create_next_step_prompt,
    create_persona_prompt,
    create_progress_ledger_prompt,
    create_summarize_result_prompt,
    create_update_hypothesis_on_stall_prompt,
    create_update_hypothesis_prompt,
    create_update_plan_on_stall_prompt,
    create_update_plan_prompt,
)


class CogenticOrchestrator(BaseGroupChatManager):
    """The CogenticOrchestrator manages a group chat with hypothesis validation."""

    def __init__(
        self,
        group_topic_type: str,
        output_topic_type: str,
        participant_topic_types: List[str],
        participant_descriptions: List[str],
        model_client: ChatCompletionClient,
        json_model_client: ChatCompletionClient | None,
        max_turns_total: int | None,
        max_turns_per_hypothesis: int | None,
        max_turns_per_test: int | None,
        max_stalls: int,
        final_answer_prompt: str,
        use_summarized_thread: bool = True,
    ):
        super().__init__(
            group_topic_type=group_topic_type,
            output_topic_type=output_topic_type,
            participant_topic_types=participant_topic_types,
            participant_descriptions=participant_descriptions,
            termination_condition=None,
            max_turns=max_turns_total,
        )
        self._model_client = model_client
        self._max_stalls = max_stalls
        self._max_turns_total = max_turns_total
        self._max_turns_per_hypothesis = max_turns_per_hypothesis
        self._max_turns_per_test = max_turns_per_test
        self._final_answer_prompt = final_answer_prompt
        self._name = "CogenticOrchestrator"
        self._max_json_retries = 10
        self._question = ""
        self._plan: CogenticPlan | None = None
        self._ledger: CogenticProgressLedger | None = None
        self._active_step: CogenticNextStep | None = None
        self._total_turns: int = 0
        self._current_hypothesis_turns: int = 0
        self._current_test_turns: int = 0
        self._current_stall_count: int = 0
        self._summarized_thread: List[AgentEvent | ChatMessage] = []
        self._use_summarized_thread = use_summarized_thread
        self.logger = logging.getLogger(TRACE_LOGGER_NAME)
        if json_model_client is None:
            self._json_model_client = json_model_client or model_client
        else:
            self._json_model_client = json_model_client

        # Create a markdown table of our team members with Name and Description
        self._team_description = "| Name | Description |\n"
        self._team_description += "| ---- | ----------- |\n"
        for topic_type, description in zip(
            self._participant_topic_types, self._participant_descriptions
        ):
            self._team_description += (
                re.sub(r"\s+", " ", f"| {topic_type} | {description} |").strip() + "\n"
            )
        self._team_description = self._team_description.strip()

    async def _publish_to_output(
        self,
        message: Any,
        cancellation_token: CancellationToken | None = None,
    ) -> None:
        """Log a message to our output topic"""

        await self.publish_message(
            GroupChatMessage(
                message=message,
            ),
            topic_id=DefaultTopicId(type=self._output_topic_type),
            cancellation_token=cancellation_token,
        )

    async def _publish_to_group(
        self,
        message: Any,
        cancellation_token: CancellationToken | None = None,
    ) -> None:
        """Log a message to our group topic"""

        await self.publish_message(
            GroupChatAgentResponse(
                agent_response=Response(chat_message=message),
            ),
            topic_id=DefaultTopicId(type=self._group_topic_type),
            cancellation_token=cancellation_token,
        )

    async def _terminate_chat(
        self, message: str, cancellation_token: CancellationToken
    ) -> None:
        """Terminate the chat.

        Args:
            message (str): The termination message.

        """
        await self.publish_message(
            message=GroupChatTermination(
                message=StopMessage(content=message, source=self._name)
            ),
            topic_id=DefaultTopicId(type=self._output_topic_type),
            cancellation_token=cancellation_token,
        )

    async def _start_chat(
        self, messages: list[ChatMessage], cancellation_token: CancellationToken
    ) -> None:
        """Start the chat."""
        await self.publish_message(
            GroupChatStart(messages=messages),
            topic_id=DefaultTopicId(type=self._group_topic_type),
            cancellation_token=cancellation_token,
        )

    async def _reset_state(self, cancellation_token: CancellationToken) -> None:
        """Reset the chat state."""
        self._current_stall_count = 0
        self._current_hypothesis_turns = 0
        self._current_test_turns = 0
        for participant_topic_type in self._participant_topic_types:
            await self._runtime.send_message(
                GroupChatReset(),
                recipient=AgentId(type=participant_topic_type, key=self.id.key),
                cancellation_token=cancellation_token,
            )

    @rpc
    async def handle_start(self, message: GroupChatStart, ctx: MessageContext) -> None:  # type: ignore
        """
        Handle the start of a group chat.

        We initialize the group chat manager and set up the initial state.

        - Gather initial evidence from the question itself
        - Create an initial plan containing hypotheses to be verified
        - Finish by selecting a hypothesis to process

        """
        assert message is not None and message.messages is not None

        # Start chat
        await self._start_chat(message.messages, ctx.cancellation_token)

        # Initialize the question by combining the initial messages
        self._question = "\n".join(
            [content_to_str(msg.content) for msg in message.messages]
        )

        # The planning conversation only exists to create a formal plan based on the question.
        # It is not broadcast to the group chat.
        planning_conversation: List[LLMMessage] = []

        self._plan = CogenticPlan()

        # Add our persona
        planning_conversation.append(SystemMessage(content=create_persona_prompt()))

        # Collect initial evidence from the question.
        planning_conversation.append(
            UserMessage(
                content=create_initial_evidence_prompt(self._question),
                source=self._name,
            )
        )
        initial_evidence = await reason_and_output_model(
            self._model_client,
            self._json_model_client,
            self._get_compatible_context(planning_conversation),
            ctx.cancellation_token,
            response_model=CogenticInitialEvidence,
            retries=self._max_json_retries,
        )
        self._plan.evidence.extend(initial_evidence.evidence)

        # Add the fact sheet to the planning conversation
        planning_conversation.append(
            AssistantMessage(
                content=initial_evidence.model_dump_markdown(title="Initial Evidence"),
                source=self._name,
            )
        )

        # Now, based on the question and the known facts, ask the model to create a plan
        planning_conversation.append(
            UserMessage(
                content=create_initial_hypotheses_prompt(self._team_description),
                source=self._name,
            )
        )
        initial_hypotheses = await reason_and_output_model(
            self._model_client,
            self._json_model_client,
            self._get_compatible_context(planning_conversation),
            ctx.cancellation_token,
            response_model=CogenticInitialHypotheses,
            retries=self._max_json_retries,
        )
        self._plan.hypotheses = initial_hypotheses.hypotheses

        await self._process_next_hypothesis(ctx.cancellation_token)

    async def _process_next_hypothesis(
        self, cancellation_token: CancellationToken
    ) -> None:
        """Process the next hypothesis in the plan.

        - Reset the agents
        - Choose the next unverified hypothesis. If none, prepare the final answer.

        """
        # Reset state
        await self._reset_state(cancellation_token)
        assert self._plan

        if not self._plan.current_hypothesis:
            raise ValueError("No current hypothesis to process. This is unexpected.")

        # Clear the ledger
        self._ledger = None

        # Clear the orchestrator message thread
        self._message_thread.clear()
        self._summarized_thread.clear()

        # Add our persona to our thread
        persona_message = TextMessage(content=create_persona_prompt(), source="system")
        self._message_thread.append(persona_message)
        self._summarized_thread.append(persona_message)

        # Create the initial message for the group chat
        current_state_message = TextMessage(
            content=create_current_state_prompt(
                question=self._question,
                team_description=self._team_description,
                plan=self._plan,
            ),
            source=self._name,
        )
        # First message in any work thread is the state
        self._message_thread.append(current_state_message)
        self._summarized_thread.append(current_state_message)

        # Publish to the output and group
        await self._publish_to_output(message=current_state_message)
        await self._publish_to_group(message=current_state_message)

        # Start the hypothesis loop
        await self._hypothesis_loop(
            cancellation_token=cancellation_token, first_iteration=True
        )

    async def _hypothesis_loop(
        self, cancellation_token: CancellationToken, first_iteration=False
    ) -> None:
        """
        This is the work loop for the current hypothesis
        After each iteration, we update facts and current hypothesis data.

        Args:
            first_iteration (bool): Whether this is the first iteration of the loop (don't create ledger).
        """
        # Check if we have reached the maximum number of turns for the orchestrator.
        if self._max_turns is not None and self._total_turns > self._max_turns:
            await self._create_final_answer(
                f"Maximum turn count reached ({self._max_turns}). Can we come to a conclusion?",
                cancellation_token,
            )
            return

        self._total_turns += 1
        self._current_hypothesis_turns += 1
        self._current_test_turns += 1

        # We exit immediately on the first iteration - no need to create a progress ledger.
        if first_iteration:
            return await self._execute_next_step(
                cancellation_token=cancellation_token,
            )

        # Request an update to the ledger
        self._ledger = await self._update_progress_ledger(
            cancellation_token=cancellation_token
        )

        assert self._plan and self._plan.current_hypothesis

        current_hypothesis = self._plan.current_hypothesis
        assert current_hypothesis.current_test

        # Updates from the ledger
        current_hypothesis.current_test.state = self._ledger.test_state.answer
        self._plan.evidence.extend(self._ledger.new_test_evidence)
        self._plan.issues.extend(self._ledger.new_issues)

        # Check if we're totally done:
        if self._ledger.original_question_answered.answer:
            # If we're marking the question as answered, the current hypothesis has been validated
            current_hypothesis.state = "verified"
            self.logger.info("Original question answered, preparing final answer...")
            await self._create_final_answer(
                self._ledger.original_question_answered.reason,
                cancellation_token,
            )
            return

        # Check if we need to replan
        if self._ledger.replan_needed.answer:
            self.logger.info("Replan requested, updating the plan...")
            return await self._replan(cancellation_token, stalled=False, requested=True)

        # Check if we're done the test/hypothesis
        if self._ledger.test_state.answer != "incomplete":
            self._current_test_turns = 0
            self.logger.info("Current test work complete.")

            # Replan on completed hypothesis
            if self._plan.current_hypothesis.all_tests_finished:
                return await self._replan(
                    cancellation_token=cancellation_token, stalled=False
                )

        # Check for stalling
        stalling = False
        if not self._ledger.forward_progress.answer:
            stalling = True
            self._current_stall_count += 1
        if self._ledger.stuck_in_loop.answer:
            stalling = True
            self._current_stall_count += 1

        if not stalling:
            # Decrement stall count if we're making progress
            self._current_stall_count = max(0, self._current_stall_count - 1)

        # Re-plan on full stall or max hypothesis turns reached
        if self._needs_replan():
            self.logger.warning(
                "Stalled or hypothesis turn count exceeded, time to update the plan."
            )
            return await self._replan(cancellation_token, stalled=True)

        # Keep going!
        await self._execute_next_step(
            cancellation_token=cancellation_token,
            next_step=self._ledger.next_step,
        )
        return

    async def _execute_next_step(
        self,
        cancellation_token: CancellationToken,
        next_step: CogenticNextStep | None = None,
    ) -> None:
        """Continue working by executing the next step.

        Args:
            next_step (CogenticNextStep | None): The next step to execute. If None, a new step will be created.
        """
        if not next_step:
            next_step = await self._create_next_step(cancellation_token)

        # Save this so we can summarize the results later
        self._active_step = next_step

        # Create the next step message and send it out to the group
        next_step_message = TextMessage(
            content=next_step.instruction_or_question.answer,
            source=self._name,
        )
        # Add it to our own internal conversation as well as our agents
        self._message_thread.append(next_step_message)
        self._summarized_thread.append(next_step_message)

        await self._publish_to_output(
            message=next_step_message, cancellation_token=cancellation_token
        )
        await self._publish_to_group(
            message=next_step_message, cancellation_token=cancellation_token
        )

        # Ask the next speaker to respond
        await self.publish_message(
            GroupChatRequestPublish(),
            topic_id=DefaultTopicId(type=next_step.next_speaker.answer),
            cancellation_token=cancellation_token,
        )

    def _needs_replan(self) -> bool:
        """Check if we need to replan based on the current state."""
        stalled = self._current_stall_count >= self._max_stalls
        hypothesis_turns_exceeded = self._max_turns_per_hypothesis is not None and (
            self._current_hypothesis_turns >= self._max_turns_per_hypothesis
        )
        test_turns_exceeded = self._max_turns_per_test is not None and (
            self._current_test_turns >= self._max_turns_per_test
        )
        return stalled or hypothesis_turns_exceeded or test_turns_exceeded

    async def _update_progress_ledger(
        self, cancellation_token: CancellationToken
    ) -> CogenticProgressLedger:
        """Update the progress ledger based on the current state of the group chat.

        Returns:
            CogenticProgressLedger: An updated progress ledger for the current state of the group chat.
        """

        # Get the active conversation (this doesn't contain previous ledger updates, just instructions/results)
        if self._use_summarized_thread:
            context = self._thread_to_context(self._summarized_thread)
        else:
            context = self._thread_to_context(self._message_thread)
        ledger_type = CogenticProgressLedger.with_speakers(
            choices=self._participant_topic_types
        )
        progress_ledger_prompt = create_progress_ledger_prompt()
        context.append(UserMessage(content=progress_ledger_prompt, source=self._name))
        progress_ledger = await reason_and_output_model(
            self._model_client,
            self._json_model_client,
            self._get_compatible_context(context),
            cancellation_token=cancellation_token,
            response_model=ledger_type,
            retries=self._max_json_retries,
        )
        self.logger.debug(f"Progress Ledger: {progress_ledger}")
        return progress_ledger

    async def _create_next_step(
        self, cancellation_token: CancellationToken
    ) -> CogenticNextStep:
        """Select the next step"""
        assert self._plan

        # Next step selection is always done as part of a conversation
        context = self._thread_to_context(self._summarized_thread)

        # Create the next step prompt
        next_step_type = CogenticNextStep.with_speaker_choices(
            choices=self._participant_topic_types
        )
        next_step_prompt = create_next_step_prompt(
            names=self._participant_topic_types,
        )
        context.append(UserMessage(content=next_step_prompt, source=self._name))
        # Get the next step
        next_step = await reason_and_output_model(
            self._model_client,
            self._json_model_client,
            self._get_compatible_context(context),
            cancellation_token=cancellation_token,
            response_model=next_step_type,
            retries=self._max_json_retries,
        )
        self.logger.debug(f"Next Step: {next_step}")

        return next_step

    async def _replan(
        self, cancellation_token: CancellationToken, stalled=False, requested=False
    ) -> None:
        """Update our plan according to the current state of the group chat.
        Args:
            stalled (bool): Whether we are doing this because we're stalled.
        """
        assert self._plan and self._plan.current_hypothesis

        persona = create_persona_prompt()
        current_state = create_current_state_prompt(
            question=self._question,
            team_description=self._team_description,
            plan=self._plan,
        )
        if stalled:
            update_hypothesis_prompt = create_update_hypothesis_on_stall_prompt()
            plan_update_prompt = create_update_plan_on_stall_prompt()
        else:
            update_hypothesis_prompt = create_update_hypothesis_prompt()
            plan_update_prompt = create_update_plan_prompt()

        messages = [
            SystemMessage(content=persona),
            UserMessage(content=current_state, source=self._name),
            UserMessage(content=update_hypothesis_prompt, source=self._name),
        ]

        hypothesis_update = await reason_and_output_model(
            self._model_client,
            self._json_model_client,
            self._get_compatible_context(messages),
            cancellation_token=cancellation_token,
            response_model=CogenticHypothesisUpdate,
            retries=self._max_json_retries,
        )
        # Add new tests to the hypothesis
        self._plan.current_hypothesis.insert_tests(hypothesis_update.new_tests)
        # Update the state of the hypothesis.
        # NOTE: if this sets the hypothesis to anything but unverified it will change future results of self._plan.current_hypothesis!
        self._plan.current_hypothesis.state = hypothesis_update.hypothesis_state.answer

        # We won't update the plan if we aren't stalled, we didn't ask for a replan, and there's work left to do
        if not stalled and not requested and self._plan.current_hypothesis:
            return await self._process_next_hypothesis(cancellation_token)

        # Otherwise, we need to update the plan
        messages = [
            SystemMessage(content=persona),
            UserMessage(content=current_state, source=self._name),
            UserMessage(content=plan_update_prompt, source=self._name),
        ]

        # Get the plan update
        plan_update = await reason_and_output_model(
            self._model_client,
            self._json_model_client,
            self._get_compatible_context(messages),
            cancellation_token=cancellation_token,
            response_model=CogenticPlanUpdate,
            retries=self._max_json_retries,
        )
        # Update the plan state
        self._plan.state = plan_update.plan_state.answer
        if self._plan.state != "in_progress":
            # If the plan is complete, we need to prepare the final answer
            self.logger.info("Plan complete, preparing final answer...")
            await self._create_final_answer(
                "No work remaining. This could mean success or failure depending on the results",
                cancellation_token,
            )
            return
        else:
            # Insert new hypotheses to the plan
            self._plan.insert_hypotheses(plan_update.new_hypotheses)
            self.logger.info("Plan still in progress, selecting next hypothesis...")
            return await self._process_next_hypothesis(cancellation_token)

    async def _create_final_answer(
        self, reason: str, cancellation_token: CancellationToken
    ) -> None:
        """Prepare the final answer for the task."""

        assert self._plan

        # Create the final answer prompt
        persona = create_persona_prompt()
        final_answer_prompt = create_final_answer_prompt(
            question=self._question,
            finish_reason=reason,
            plan=self._plan,
        )
        messages = [
            SystemMessage(content=persona),
            UserMessage(content=final_answer_prompt, source=self._name),
        ]

        final_answer = await reason_and_output_model(
            self._model_client,
            self._json_model_client,
            messages=self._get_compatible_context(messages),
            cancellation_token=cancellation_token,
            response_model=CogenticFinalAnswer,
            retries=self._max_json_retries,
        )

        message = TextMessage(
            content=final_answer.model_dump_markdown(), source=self._name
        )
        self._message_thread.append(message)
        self._summarized_thread.append(message)

        # Publish the response message
        await self._publish_to_output(
            message=message,
            cancellation_token=cancellation_token,
        )
        await self._publish_to_group(
            message=message,
            cancellation_token=cancellation_token,
        )

        # Terminate
        await self._terminate_chat(
            message=reason,
            cancellation_token=cancellation_token,
        )

    @event
    async def handle_agent_response(
        self, message: GroupChatAgentResponse, ctx: MessageContext
    ) -> None:
        """Handle the response from an agent in our group chat."""
        # Add this message to our ongoing work thread
        self._message_thread.append(message.agent_response.chat_message)
        # Summarize what happened for our plan history
        await self._summarize_action(
            message.agent_response.chat_message, ctx.cancellation_token
        )
        await self._hypothesis_loop(ctx.cancellation_token)

    async def _summarize_action(
        self, message: ChatMessage, cancellation_token: CancellationToken
    ):
        """Summarize the action we just took

        Args:
            message (GroupChatAgentResponse): The response from the agent.
            cancellation_token (CancellationToken): The cancellation token for the operation.

        """

        # Join the content into a single string
        content_str = content_to_str(message.content)

        assert self._active_step
        assert self._plan
        assert self._plan.current_hypothesis
        assert self._plan.current_hypothesis.current_test

        # Get the action summary
        action_summary_prompt = create_summarize_result_prompt(
            response=content_str,
        )
        action_summary_response = await self._json_model_client.create(
            messages=[
                UserMessage(content=action_summary_prompt, source=self._name),
            ],
            cancellation_token=cancellation_token,
        )
        assert isinstance(action_summary_response.content, str)
        action_summary = action_summary_response.content
        # Create the action
        action = CogenticAction(
            goal=self._active_step.goal.answer,
            team_member_name=self._active_step.next_speaker.answer,
            test_name=self._plan.current_hypothesis.current_test.name,
            outcome=action_summary,
        )
        # Add the action to the plan
        self._plan.actions.append(action)
        # Store the result in our summarized thread:
        self._summarized_thread.append(
            TextMessage(
                content=action_summary,
                source=self._active_step.next_speaker.answer,
            )
        )

    def _thread_to_context(
        self, thread: list[AgentEvent | ChatMessage]
    ) -> List[LLMMessage]:
        """Convert the message thread to a context for the model."""
        context: List[LLMMessage] = []
        for m in thread:
            if isinstance(m, ToolCallRequestEvent | ToolCallExecutionEvent):
                # Ignore tool call messages.
                continue
            elif isinstance(m, StopMessage | HandoffMessage):
                context.append(UserMessage(content=m.content, source=m.source))
            elif m.source == self._name:
                assert isinstance(m, TextMessage | ToolCallSummaryMessage)
                context.append(AssistantMessage(content=m.content, source=m.source))
            else:
                assert isinstance(
                    m, (TextMessage, MultiModalMessage, ToolCallSummaryMessage)
                )
                context.append(UserMessage(content=m.content, source=m.source))
        return context

    def _get_compatible_context(self, messages: List[LLMMessage]) -> List[LLMMessage]:
        """Ensure that the messages are compatible with the underlying client, by removing images if needed."""
        if self._model_client.model_info["vision"]:
            return messages
        else:
            return remove_images(messages)

    async def validate_group_state(self, messages: List[ChatMessage] | None) -> None:
        pass

    async def save_state(self) -> Mapping[str, Any]:
        state = CogenticState(
            message_thread=list(self._message_thread),
            current_turn=self._current_turn,
            question=self._question,
            plan=self._plan,
            total_turns=self._total_turns,
            stalls=self._current_stall_count,
        )
        return state.model_dump()

    async def load_state(self, state: Mapping[str, Any]) -> None:
        orchestrator_state = CogenticState.model_validate(state)
        self._message_thread = orchestrator_state.message_thread
        self._current_turn = orchestrator_state.current_turn
        self._question = orchestrator_state.question
        self._plan = orchestrator_state.plan
        self._total_turns = orchestrator_state.total_turns
        self._current_stall_count = orchestrator_state.stalls

    async def select_speaker(self, thread: List[AgentEvent | ChatMessage]) -> str:
        """Not used in this orchestrator, we select next speaker in _orchestrate_step."""
        return ""

    async def reset(self) -> None:
        """Reset the group chat manager."""
        self._message_thread.clear()
        self._total_turns = 0
        self._current_stall_count = 0
        self._question = ""
        self._plan = None
        self._ledger = None
        self._current_hypothesis_turns = 0
        self._current_test_turns = 0



from __future__ import annotations

from typing import Literal

from pydantic import Field

from cogentic.orchestration.models.base import CogenticBaseModel
from cogentic.orchestration.models.test import CogenticTest

CogenticHypothesisState = Literal[
    "unverified",
    "verified",
    "unverifiable",
]


class CogenticHypothesis(CogenticBaseModel):
    """Hypothesis for the cogentic system."""

    hypothesis: str = Field(description="Hypothesis to be tested")
    state: CogenticHypothesisState = Field(description="State of the hypothesis")
    completion_summary: str | None = Field(
        description="When completed, a summary of the results"
    )
    tests: list[CogenticTest] = Field(
        description="Tests for the hypothesis. Hypotheses must have at least one test",
        min_length=1,
    )

    @property
    def all_tests_finished(self) -> bool:
        """Check if all tests are completed or we're in a completed state."""
        return all(test.state != "incomplete" for test in self.tests)

    @property
    def all_tests_completed(self) -> bool:
        """Check if all tests are completed."""
        return all(test.state == "complete" for test in self.tests)

    @property
    def current_test(self) -> CogenticTest | None:
        """Get the current test to be completed."""
        for test in self.tests:
            if test.state == "incomplete":
                return test
        return None

    def insert_tests(self, to_insert: CogenticTest | list[CogenticTest]) -> None:
        """Insert test(s) into the hypothesis in front of our current test."""
        new_tests = to_insert if isinstance(to_insert, list) else [to_insert]
        if self.current_test:
            index = self.tests.index(self.current_test)
            self.tests[index:index] = new_tests
        else:
            self.tests.extend(new_tests)


class CogenticInitialHypotheses(CogenticBaseModel):
    """Initial hypotheses for the cogentic system."""

    hypotheses: list[CogenticHypothesis] = Field(
        description="Hypotheses to be tested", min_length=1
    )



from typing import Literal

from pydantic import Field

from cogentic.orchestration.models.action import CogenticAction
from cogentic.orchestration.models.artifact import CogenticArtifact
from cogentic.orchestration.models.base import CogenticBaseModel
from cogentic.orchestration.models.evidence import CogenticEvidence
from cogentic.orchestration.models.hypothesis import CogenticHypothesis
from cogentic.orchestration.models.issue import CogenticIssue

CogenticPlanState = Literal[
    "in_progress",
    "completed",
    "failed",
]


class CogenticPlan(CogenticBaseModel):
    """Plan for the cogentic system."""

    state: CogenticPlanState = Field(
        description="State of the plan",
        default="in_progress",
    )
    hypotheses: list[CogenticHypothesis] = Field(
        description="Hypotheses to be tested", default_factory=list
    )
    artifacts: list[CogenticArtifact] = Field(
        description="Artifacts we have created as part of our work e.g., scripts, files, etc.",
        default_factory=list,
    )
    issues: list[CogenticIssue] = Field(
        description="Issues we have encountered as part of our work e.g., bugs, errors, etc.",
        default_factory=list,
    )
    evidence: list[CogenticEvidence] = Field(
        description="Evidence we have collected as part of our work e.g., sources, content, etc.",
        default_factory=list,
    )
    actions: list[CogenticAction] = Field(
        description="History of all actions taken and their outcomes as part of this plan.",
        default_factory=list,
    )

    @property
    def current_hypothesis(self) -> CogenticHypothesis | None:
        """Get the current hypothesis to be tested."""
        for hypothesis in self.hypotheses:
            if hypothesis.state == "unverified":
                return hypothesis
        return None

    def insert_hypotheses(
        self, to_insert: CogenticHypothesis | list[CogenticHypothesis]
    ) -> None:
        """Insert a hypothesis into the plan in front of our current hypothesis."""
        new_hypotheses = to_insert if isinstance(to_insert, list) else [to_insert]
        if self.current_hypothesis:
            index = self.hypotheses.index(self.current_hypothesis)
            self.hypotheses[index:index] = new_hypotheses
        else:
            self.hypotheses.extend(new_hypotheses)



from typing import Literal

from pydantic import Field

from cogentic.orchestration.models.base import CogenticBaseModel


class CogenticIssue(CogenticBaseModel):
    """Issue for the cogentic system."""

    name: str = Field(description="Name of the issue")
    description: str = Field(description="Description of the issue")
    severity: Literal["low", "medium", "high"] = Field(
        description="Severity of the issue"
    )


class CogenticIssueContainer(CogenticBaseModel):
    """Container for issues."""

    entries: list[CogenticIssue] = Field(
        description="List of issues encountered during the work",
    )



import json
import re
from typing import Type, TypeVar

from autogen_core import CancellationToken
from autogen_core.models import (
    AssistantMessage,
    ChatCompletionClient,
    LLMMessage,
    UserMessage,
)
from pydantic import BaseModel, ValidationError

T = TypeVar("T", bound=BaseModel)

FORMAT_PROMPT = """\

### Response Output Schema

Now, I need you to format your previous response in a specific format.

Here is the SCHEMA of the response format:

```json
{response_schema}
```

- Note that you need to create an instance of the model that ADHERES to this schema; not the schema itself.


### Note

Please output your response in a json-formatted code block adhering to the schema. Make sure to wrap your json in markdown tags, e.g.:

```json
... your json content here ...
```

"""

RETRY_MESSAGE = """\

## Response Format Error

- We were unable to parse the JSON output from your response. 
- The output was not in the expected format. 

### Here was your input:

```json
{input}
```

### Error

The error was:

```text
{error}
```

- Consider whether you accidentally output an instance of the schema itself instead of a json object that adheres to the schema.
- Please try to adjust the JSON and respond correctly.
"""


class CogenticOutputParsingError(Exception):
    """Exception raised for errors in the output parsing."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message


def _extract_json_from_response(response: str) -> str | None:
    json_match = re.search(r"```json\n(.*?)\n```", response, re.DOTALL)
    if json_match:
        json_str = json_match.group(1)
        # Remove any leading/trailing whitespace
        json_str = json_str.strip()
        return json_str
    return None


async def _reason_and_request_model_via_markdown(
    model_client: ChatCompletionClient,
    json_model_client: ChatCompletionClient,
    messages: list[LLMMessage],
    cancellation_token: CancellationToken,
    response_model: Type[T],
    retries: int = 3,
):
    errors = []
    # First, get the model to respond using the original prompt
    first_response = await model_client.create(
        messages=messages,
        cancellation_token=cancellation_token,
    )
    assert isinstance(first_response.content, str)
    create_messages = messages[:] + [
        AssistantMessage(content=first_response.content, source="assistant"),
        UserMessage(
            content=FORMAT_PROMPT.format(
                response_schema=json.dumps(response_model.model_json_schema(), indent=2)
            ),
            source="assistant",
        ),
    ]
    retry_messages = create_messages[:]
    for _ in range(retries):
        try:
            response = await json_model_client.create(
                messages=retry_messages,
                cancellation_token=cancellation_token,
            )
            assert isinstance(response.content, str)
            json_object = None
            try:
                json_content = _extract_json_from_response(response.content)
                if json_content is None:
                    raise CogenticOutputParsingError(
                        "No JSON markdown block found in the response. Please ensure your response is formatted correctly."
                    )
                json_object = json.loads(json_content)
                return response_model.model_validate(json_object)
            except (CogenticOutputParsingError, ValidationError) as e:
                # We don't want to include multiple error messages in the retry
                input_str = (
                    json.dumps(json_object, indent=2)
                    if json_object
                    else response.content
                )
                retry_messages = create_messages[:]
                retry_message = RETRY_MESSAGE.format(
                    input=input_str,
                    error=str(e),
                )
                errors.append(e)
                retry_messages.append(UserMessage(content=retry_message, source="user"))

        except Exception as e:
            errors.append(e)
            # We don't want to include multiple error messages in the retry
            retry_messages = create_messages[:]
            retry_message = f"Unexpected error. Please try again.\n\nError: {e}"
            retry_messages.append(UserMessage(content=retry_message, source="user"))

    raise ValueError(
        f"Failed to get a valid response after multiple attempts:\n{errors}"
    )


async def _reason_and_request_model_directly(
    model_client: ChatCompletionClient,
    json_model_client: ChatCompletionClient,
    messages: list[LLMMessage],
    cancellation_token: CancellationToken,
    response_model: Type[T],
    retries: int = 3,
) -> T:
    # First, get the model to respond using the original prompt
    first_response = await model_client.create(
        messages=messages,
        cancellation_token=cancellation_token,
    )
    assert isinstance(first_response.content, str)
    create_messages = messages + [
        AssistantMessage(content=first_response.content, source="assistant"),
        UserMessage(
            content="Please format your response using the provided format",
            source="user",
        ),
    ]
    errors = []
    retry_messages = create_messages[:]
    for _ in range(retries):
        try:
            # Now use the json model client to get the JSON output
            model_response = await json_model_client.create(
                messages=retry_messages,
                cancellation_token=cancellation_token,
                extra_create_args={"response_format": response_model},
            )
            assert isinstance(model_response.content, str)
            return response_model.model_validate_json(model_response.content)
        except Exception as e:
            # We don't want to include multiple error messages in the retry
            retry_messages = create_messages[:]
            retry_message = RETRY_MESSAGE.format(
                error=e,
            )
            retry_messages.append(UserMessage(content=retry_message, source="user"))
            errors.append(e)
            continue
    raise ValueError(
        f"Failed to get a valid response after multiple attempts:\n{errors}"
    )


async def reason_and_output_model(
    model_client: ChatCompletionClient,
    json_model_client: ChatCompletionClient,
    messages: list[LLMMessage],
    cancellation_token: CancellationToken,
    response_model: Type[T],
    retries: int = 3,
) -> T:
    """
    Reason and output the model.

    Args:
        model_client (ChatCompletionClient): The model client used for reasoning/inference
        json_model_client (ChatCompletionClient): The model client used to request/extract JSON output
        messages (list[LLMMessage]): The messages to send to the model
        cancellation_token (CancellationToken): The cancellation token to use for the request
        response_model (Type[T]): The model to use for the response
        retries (int, optional): The number of retries to attempt. Defaults to 3.

    Returns:
        T: The model output type
    """
    return await _reason_and_request_model_via_markdown(
        model_client=model_client,
        json_model_client=json_model_client,
        messages=messages,
        cancellation_token=cancellation_token,
        response_model=response_model,
        retries=retries,
    )



from typing import Generic, TypeVar

from pydantic import Field

from cogentic.orchestration.models.base import CogenticBaseModel

T = TypeVar("T")


class CogenticReasonedStringAnswer(CogenticBaseModel):
    """Reasoned answer for the progress ledger."""

    reason: str = Field(description="Reason for the answer")
    answer: str = Field(description="Answer to the question")


class CogenticReasonedBooleanAnswer(CogenticBaseModel):
    """Reasoned answer for the progress ledger."""

    reason: str = Field(description="Reason for the answer")
    answer: bool = Field(description="Answer to the question")


class CogenticReasonedChoiceAnswer(CogenticBaseModel, Generic[T]):
    """Reasoned answer for the progress ledger from a set of choices."""

    reason: str = Field(description="Reason for the answer")
    answer: T = Field(
        description="Answer to the question.",
    )



from typing import Union

from pydantic import Field

from cogentic.orchestration.models.base import CogenticBaseModel


class CogenticQuestionEvidence(CogenticBaseModel):
    """Evidence for the cogentic system."""

    description: str = Field(
        description="Description of the evidence, including it's relevance to the question"
    )
    content: str = Field(description="Relevant content from the source")


class CogenticTestEvidence(CogenticBaseModel):
    """Evidence for the cogentic system."""

    test_name: str = Field(description="Name of the test which produced this evidence")
    team_member_name: str = Field(
        description="Name of the team member who produced this evidence during their work on the test"
    )
    content: str = Field(description="The evidence itself")


CogenticEvidence = Union[CogenticQuestionEvidence, CogenticTestEvidence]


class CogenticEvidenceContainer(CogenticBaseModel):
    """Container for evidence."""

    entries: list[CogenticEvidence] = Field(
        description="List of evidence collected from the question or test",
    )


class CogenticInitialEvidence(CogenticBaseModel):
    """Initial evidence for the cogentic system."""

    evidence: list[CogenticQuestionEvidence] = Field(
        description="List of evidence collected from the question",
    )



# import json
from pathlib import Path

from cogentic.orchestration.models.plan import CogenticPlan

PROMPTS_DIR = Path(__file__).parent

PERSONA_PROMPT_PATH = PROMPTS_DIR / "persona.md"
PERSONA_PROMPT = PERSONA_PROMPT_PATH.read_text()


def create_persona_prompt() -> str:
    return PERSONA_PROMPT


INITIAL_EVIDENCE_PROMPT_PATH = PROMPTS_DIR / "initial_evidence.md"
INITIAL_EVIDENCE_PROMPT = INITIAL_EVIDENCE_PROMPT_PATH.read_text()


def create_initial_evidence_prompt(
    question: str,
) -> str:
    return INITIAL_EVIDENCE_PROMPT.format(
        question=question,
    )


INITIAL_HYPOTHESES_PROMPT_PATH = PROMPTS_DIR / "initial_hypotheses.md"
INITIAL_HYPOTHESES_PROMPT = INITIAL_HYPOTHESES_PROMPT_PATH.read_text()


def create_initial_hypotheses_prompt(team_description: str) -> str:
    return INITIAL_HYPOTHESES_PROMPT.format(
        team_description=team_description,
    )


CURRENT_STATE_PROMPT_PATH = PROMPTS_DIR / "current_state.md"
CURRENT_STATE_PROMPT = CURRENT_STATE_PROMPT_PATH.read_text()


def create_current_state_prompt(
    question: str,
    team_description: str,
    plan: CogenticPlan,
) -> str:
    current_test = "No tests have work remaining."
    current_hypothesis = "No hypotheses have work remaining."
    if plan.current_hypothesis:
        if plan.current_hypothesis.current_test:
            current_test = plan.current_hypothesis.current_test.model_dump_markdown()
        current_hypothesis = plan.current_hypothesis.model_dump_markdown()
    return CURRENT_STATE_PROMPT.format(
        question=question,
        current_hypothesis=current_hypothesis,
        current_test=current_test,
        team_description=team_description,
        evidence=plan.model_dump_field_as_markdown("evidence"),
        issues=plan.model_dump_field_as_markdown("issues"),
    )


CREATE_PROGRESS_LEDGER_PROMPT_PATH = PROMPTS_DIR / "create_progress_ledger.md"
CREATE_PROGRESS_LEDGER_PROMPT = CREATE_PROGRESS_LEDGER_PROMPT_PATH.read_text()


def create_progress_ledger_prompt() -> str:
    return CREATE_PROGRESS_LEDGER_PROMPT


FINAL_ANSWER_PROMPT_PATH = PROMPTS_DIR / "final_answer.md"
FINAL_ANSWER_PROMPT = FINAL_ANSWER_PROMPT_PATH.read_text()


def create_final_answer_prompt(
    question: str,
    finish_reason: str,
    plan: CogenticPlan,
) -> str:
    return FINAL_ANSWER_PROMPT.format(
        question=question,
        finish_reason=finish_reason,
        plan=plan.model_dump_markdown(),
    )


UPDATE_HYPOTHESIS_PATH = PROMPTS_DIR / "update_hypothesis.md"
UPDATE_HYPOTHESIS_PROMPT = UPDATE_HYPOTHESIS_PATH.read_text()


def create_update_hypothesis_prompt() -> str:
    return UPDATE_HYPOTHESIS_PROMPT


UPDATE_PLAN_PATH = PROMPTS_DIR / "update_plan.md"
UPDATE_PLAN_PROMPT = UPDATE_PLAN_PATH.read_text()


def create_update_plan_prompt() -> str:
    """Update the plan when we have completed work on a hypothesis."""
    return UPDATE_PLAN_PROMPT


UPDATE_HYPOTHESIS_ON_STALL_PROMPT_PATH = PROMPTS_DIR / "update_hypothesis_on_stall.md"
UPDATE_HYPOTHESIS_ON_STALL_PROMPT = UPDATE_HYPOTHESIS_ON_STALL_PROMPT_PATH.read_text()


def create_update_hypothesis_on_stall_prompt() -> str:
    return UPDATE_HYPOTHESIS_ON_STALL_PROMPT


UPDATE_PLAN_ON_STALL_PROMPT_PATH = PROMPTS_DIR / "update_plan_on_stall.md"
UPDATE_PLAN_ON_STALL_PROMPT = UPDATE_PLAN_ON_STALL_PROMPT_PATH.read_text()


def create_update_plan_on_stall_prompt() -> str:
    """Update the plan when we have stalled on a hypothesis."""
    return UPDATE_PLAN_ON_STALL_PROMPT


NEXT_STEP_PROMPT_PATH = PROMPTS_DIR / "next_step.md"
NEXT_STEP_PROMPT = NEXT_STEP_PROMPT_PATH.read_text()


def create_next_step_prompt(
    names: list[str],
) -> str:
    return NEXT_STEP_PROMPT.format(
        names=", ".join(names),
    )


SUMMARIZE_RESULT_PROMPT_PATH = PROMPTS_DIR / "summarize_result.md"
SUMMARIZE_RESULT_PROMPT = SUMMARIZE_RESULT_PROMPT_PATH.read_text()


def create_summarize_result_prompt(
    response: str,
) -> str:
    response = response or "No response provided."

    return SUMMARIZE_RESULT_PROMPT.format(
        response=response,
    )



from typing import Self, Type

from pydantic import Field, model_validator

from cogentic.orchestration.models.base import CogenticBaseModel
from cogentic.orchestration.models.evidence import CogenticTestEvidence
from cogentic.orchestration.models.issue import CogenticIssue
from cogentic.orchestration.models.orchestration import CogenticNextStep
from cogentic.orchestration.models.reasoning import (
    CogenticReasonedBooleanAnswer,
    CogenticReasonedChoiceAnswer,
)
from cogentic.orchestration.models.test import CogenticTestState


class CogenticProgressLedger(CogenticBaseModel):
    """Progress ledger for the cogentic system."""

    original_question_answered: CogenticReasonedBooleanAnswer = Field(
        description="Is the original question fully answered?",
    )
    test_state: CogenticReasonedChoiceAnswer[CogenticTestState] = Field(
        description="What is the state of the current test? Have we completed it? Are we stuck, and need to abandon it? Or are we still working on it?",
    )
    replan_needed: CogenticReasonedBooleanAnswer = Field(
        description="Based on our most recent results, do we need to adjust course? This could include changing the test plan or creating a new hypothesis.",
    )
    new_test_evidence: list[CogenticTestEvidence] = Field(
        description="Include any new evidence from our work on the current test.",
    )
    stuck_in_loop: CogenticReasonedBooleanAnswer = Field(
        description="Are we in a loop where we keep repeating the same request without making progress? ",
    )
    forward_progress: CogenticReasonedBooleanAnswer = Field(
        description="Are we making forward progress?",
    )
    new_issues: list[CogenticIssue] = Field(
        description="Include any new issues we have encountered.",
    )
    next_step: CogenticNextStep | None = Field(
        description="The next step to take. This can be null if we're marking the original question as answered or requesting a replan.",
    )

    @classmethod
    def with_speakers(cls, choices: list[str]) -> Type["CogenticProgressLedger"]:
        """Create a new type from our class, where the next speaker is limited to a set of choices."""
        # Create the choice type with proper annotation
        next_step_type = CogenticNextStep.with_speaker_choices(choices)

        return type(
            "CogenticProgressLedgerWithSpeakers",
            (cls,),
            {
                "__annotations__": {"next_step": next_step_type | None},
            },
        )

    @model_validator(mode="after")
    def validate_next_step(self) -> Self:
        """Validate that the next step is not None if the original question is not answered."""
        if self.next_step is None and self.original_question_answered is False:
            raise ValueError(
                "If we're not marking the original question as answered, we must have a next step!",
            )
        return self
